;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "pr-review" "0.2.0"
  "Review github PR."
  '((emacs         "27.1")
    (magit-section "4.0")
    (magit         "4.0")
    (markdown-mode "2.5")
    (ghub          "3.5"))
  :url "https://github.com/blahgeek/emacs-pr-review"
  :commit "7c2ce9deafe5158b4b0fe7c1e70104d64727c15e"
  :revdesc "7c2ce9deafe5"
  :keywords '("tools")
  :authors '(("Yikai Zhao" . "yikai@z1k.dev"))
  :maintainers '(("Yikai Zhao" . "yikai@z1k.dev")))
