query($repo_owner:String!, $repo_name:String!, $cursor:String) { 
  repository(owner: $repo_owner, name: $repo_name) {
    assignableUsers(first: 100, after: $cursor) {
      pageInfo {
        endCursor
        hasNextPage
      }
      nodes {
        id
        login
        name
      }
    }
  }
}